/***************************************************************************
 *  Copyright:http://www.naradabrokering.org/index.html
 *  Author: Jaliya Ekanayake
 *  Email : jaliyae@gmail.com
 ****************************************************************************/
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/bn.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "UserRSA.h"

#include "AgentServiceClient.h"
#include "AgentConfig.h"
#include "ServiceException.h"
#include "ClientCall.h"
#include "MessageCreator.h"
#include "constants.h"
#include "ServiceDiscoveryResponse.h"
#include "EncryptedCoverMessage.h"


AgentServiceClient::~AgentServiceClient(){
	/*clean the singleton object of UserRSA*/
	this->rsa->unref();
}

AgentServiceClient::AgentServiceClient(string configFile):configFile(configFile)
{								
}

void 
AgentServiceClient::init()
{
	AgentConfig cfg(configFile);
	try{	
		cfg.getConfiguration(&ctx);	
	}catch(ServiceException& e){
		throw("Coud not initialize the AgentServiceClient "+e.getMessage());
	}
	
	//TODO:Remove this.
	this->ctx.display();		
	
	rsa=UserRSA::getUserRSA();
	 
	/*Initialize SecurityUtil*/
    this->secUtil.init();
	 
	this->call.init(&ctx);     
    this->msgCreator.init(&ctx,rsa,&secUtil);	
       
	
}
/*
 * Returns the first AgentDiscoveryResponse received.
 * User should delete the returned AgentDiscoveryResponse as
 * it is allocated in the heap. 
 * */
AgentDiscoveryResponse * 
AgentServiceClient::discoverAgentFIFO()
{ 
  	/*
  	 * Need a UUID Generator for C++ to generate unique topics 
  	 * Still could not find a library which does this. So going
  	 * ahead with timestamps.
  	 * */  	 
	 
	string subTopic = DISCOVER_AGENT + "/" +Utils::getCurrentTimeMicroAsString();	
	
	NBCEnvelope env;	
	try{
		this->msgCreator.createDiscoverAgentRequest(subTopic,&env);
	}catch(ServiceException& e){
		throw ServiceException("Error while creating messages. "+e.getMessage());
	}	
	
	char buff[env.getSerializedLength()];
	int env_length=0;
	env.getBytes(buff,&env_length);
	
	/*Sanity Check*/
	assert(env.getSerializedLength()>=env_length);
	
	int num_return=0;	
	char **responses = new char * [ONE_RESPONSE];
	
	//FIFO, So wait for only one response
	this->call.sendReceive(DISCOVER_AGENT,subTopic,buff,env_length,ONE_RESPONSE,responses,&num_return);	
	
	if(num_return>0){		
		NBCCoverMessage *coverMsgs[num_return];
			
		int ret=getCoverMessages(responses,num_return,coverMsgs);		
		
		/*Delete the response array*/
		for(int i =0;i<ONE_RESPONSE;i++){
			/*These are the payloads of the nbevents received*/
			delete [] (responses[i]);
		}
		delete[] responses;		
		
		
		AgentDiscoveryResponse * res=NULL;
		if(ret>0){
		 res=selectAgentFIFO(coverMsgs,num_return);		 
		}else{
			throw ServiceException("Could not discover any agent...");
		} 
		
		if(res==NULL){
			/*Delete the cover messages*/
			for(int i =0;i<ONE_RESPONSE;i++){
				delete coverMsgs[i];
			}
			throw ServiceException("Could not discover any agent...");
		}
		
		/*Delete the cover messages*/
			for(int i =0;i<ONE_RESPONSE;i++){
				delete coverMsgs[i];
		}
		return res;		
	}else{
		/*Delete the response array*/
		delete[] responses;
		throw ServiceException("Could not discover any agent...");
	}	
}

/*
 * Returns the first NBCCoverMessage which is not NULL.
 * If all of them are NULL then return NULL
 * */
AgentDiscoveryResponse * 
AgentServiceClient::selectAgentFIFO(NBCCoverMessage *coverMsgs[],int numReturned){
	AgentDiscoveryResponse* res=NULL;
	for(int i=0;i<numReturned;i++){
		res=(AgentDiscoveryResponse *)coverMsgs[i]->getPayloadMessage();
		if(res!=NULL){
			break;
		}
	}
	return res;	
}


int		
AgentServiceClient::getCoverMessages(char **responses,int num_res,NBCCoverMessage *coverMsgs[]){
	int numCoverMsgs=0;
	for(int i=0;i<num_res;i++){		
		/*
		 * First get the certificate and verify the signature.
		 * */
		NBCEnvelope env;
		env.fromBytes(responses[i]);	
		
		if(this->secUtil.verifySignature(env)){
			numCoverMsgs++;
		    if(env.isSecure()){    	
    	       	EncryptedCoverMessage encCoverMsg;
    	       	encCoverMsg.fromBytes(env.getPayload()); 
    	       	
        	 
    	       	unsigned char secKeyDecBuf[RSA_size((const RSA*)rsa->getPrivateRSA())];    	       	
    	       	int decLength=RSA_private_decrypt(encCoverMsg.getSecretKeyLength(),(unsigned char *)encCoverMsg.getSecretKey(),secKeyDecBuf,(RSA*)rsa->getPrivateRSA(),RSA_PKCS1_PADDING);
    	       	string secKey((char *)secKeyDecBuf,decLength);
    	       	
    	       	int encMsgLength=encCoverMsg.getMessageLength();
    	       	
    	        unsigned char decBuf[encMsgLength+AES_INC];
			  	bzero (& decBuf, encMsgLength+AES_INC);
				int msgLength=0;
    	       	
    	       	if(!secUtil.decryptAES((const unsigned char*)encCoverMsg.getMessage(),encMsgLength,decBuf,msgLength,secKey)){
    	       		//TODO Cleanup
    	       		cerr<<"Decrypting CoverMessage Falied"<<endl;
    	       		numCoverMsgs--;
    	       		continue;
    	       	}     	
    	       	
    	        NBCCoverMessage *coverMsg=new NBCCoverMessage();         	  
                coverMsg->fromBytes((char *)decBuf);               
                coverMsgs[i]= coverMsg; 	       	
    	       	cout<<"Handled a secure cover message "<<endl;    	       	
            }else{
         	   NBCCoverMessage *coverMsg=new NBCCoverMessage();         	  
               coverMsg->fromBytes(env.getPayload());               
               coverMsgs[i]= coverMsg;              
            }                
		}else{
			/*Can be only for some messages, let's log it and proceed.*/
			cerr<<"Signature Verification Failed"<<endl;
		}		
	}
	return 	numCoverMsgs;	
}

vector<string> 
AgentServiceClient::discoverServices(AgentDiscoveryResponse *discoveryResponse){
	
	/*
  	 * Need a UUID Generator for C++ to generate unique topics 
  	 * Still could not find a library which does this. So going
  	 * ahead with timestamps.
  	 * */	 
	string subTopic = CLIENT_TEMP_RESPONSE_TPOIC + "/" +Utils::getCurrentTimeMicroAsString();	
	string pubTopic=DISCOVER_ROOTLET+"/"+discoveryResponse->getAgentId();
	
	NBCEnvelope env;
	X509 *cert;
	discoveryResponse->getAgentCertificate(&cert);	
	try{
		this->msgCreator.createDiscoverServiceRequest(subTopic,&env,cert);
	}catch(ServiceException& e){
		throw ServiceException("Error while creating messages. "+e.getMessage());
	}	
	
	char buff[env.getSerializedLength()];
	int env_length=0;
	env.getBytes(buff,&env_length);
	
	/*Sanity Check*/
	assert(env.getSerializedLength()>=env_length);
	
	int num_return=0;	
	char **responses = new char * [ONE_RESPONSE];
	
	//FIFO, So wait for only one response
	cout<<"Sending Service Discovery to this topic = "<<DISCOVER_ROOTLET+discoveryResponse->getAgentId()<<endl;
	this->call.sendReceive(pubTopic,subTopic,buff,env_length,ONE_RESPONSE,responses,&num_return);	
	
	if(num_return>0){		
		NBCCoverMessage *coverMsgs[num_return];			
		int ret=getCoverMessages(responses,num_return,coverMsgs);		
		
		/*Delete the response array*/
		for(int i =0;i<ONE_RESPONSE;i++){
			/*These are the payloads of the nbevents received*/
			delete [] (responses[i]);
		}
		delete[] responses;	
		
		if(ret>0){		
			ServiceDiscoveryResponse *res=(ServiceDiscoveryResponse *)coverMsgs[0]->getPayloadMessage();
			if(res!=NULL){
				vector<string> services=res->getServiceDescriptions();
				delete res;
				if(services.size()==0){
					for(int i =0;i<ONE_RESPONSE;i++){
						delete coverMsgs[i];
					}
					throw ServiceException("Could not discover any service...");
				}else{
					return services;
				}
			}else{
				//Delete the cover messages
				for(int i =0;i<ONE_RESPONSE;i++){
					delete coverMsgs[i];
				}
				throw ServiceException("Could not discover any service...");
			}			
		}else{
			//Delete the cover messages
			for(int i =0;i<ONE_RESPONSE;i++){
					delete coverMsgs[i];
			}
			throw ServiceException("Could not discover any service...");
		}
	
//	vector<string> services;
//	services.push_back("services1");
//	services.push_back("services2");
//	services.push_back("services3");
//	services.push_back("services4");
//	services.push_back("services5");
//	
//	return services;
		
		
		
			
	}else{
		//Delete the response array
		throw ServiceException("Could not discover any service...");
	}	
	
	
////	

}



